@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # This file is part of Xpra.
 # Copyright (C) 2009-2020 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
@@ -24,22 +24,21 @@ from gi.repository import Pango, GLib, Gtk, GdkPixbuf
 
 from xpra.gtk_common.gobject_compat import register_os_signals
 from xpra.scripts.config import read_config, make_defaults_struct, validate_config, save_config
-from xpra.codecs.codec_constants import PREFERED_ENCODING_ORDER
-from xpra.gtk_common.quit import gtk_main_quit_really
+from xpra.codecs.codec_constants import PREFERRED_ENCODING_ORDER
 from xpra.gtk_common.gtk_util import (
     add_close_accel, scaled_image, color_parse,
-    choose_file, imagebutton,
+    choose_file, imagebutton, get_icon_pixbuf,
     )
 from xpra.util import DEFAULT_PORT, csv, repr_ellipsized
 from xpra.os_util import WIN32, OSX, POSIX
 from xpra.make_thread import start_thread
-from xpra.client.gtk_base.gtk_tray_menu_base import make_encodingsmenu
+from xpra.client.gtk_base.menu_helper import make_encodingsmenu
 from xpra.gtk_common.about import about
 from xpra.scripts.main import (
-    connect_to, make_client, configure_network, is_local,
+    connect_to, make_client, is_local,
     add_ssh_args, parse_ssh_string, add_ssh_proxy_args,
+    configure_network, configure_env, configure_logging,
     )
-from xpra.platform.paths import get_icon_dir
 from xpra.platform import get_username
 from xpra.log import Logger, enable_debug_for
 
@@ -119,12 +118,6 @@ def has_mdns():
 def noop(*args):
     log("noop%s", args)
 
-def get_pixbuf(icon_name):
-    icon_filename = os.path.join(get_icon_dir(), icon_name)
-    if os.path.exists(icon_filename):
-        return GdkPixbuf.Pixbuf.new_from_file(icon_filename)
-    return None
-
 
 class ApplicationWindow:
 
@@ -192,7 +185,7 @@ class ApplicationWindow:
         self.window.set_position(Gtk.WindowPosition.CENTER)
         self.window.set_wmclass("xpra-launcher-gui", "Xpra-Launcher-GUI")
         add_close_accel(self.window, self.destroy)
-        icon = get_pixbuf("connect.png")
+        icon = get_icon_pixbuf("connect.png")
         if icon:
             self.window.set_icon(icon)
 
@@ -329,13 +322,13 @@ class ApplicationWindow:
         self.username_entry.set_width_chars(16)
         self.username_entry.connect("changed", self.validate)
         self.username_entry.connect("activate", self.connect_clicked)
-        self.username_entry.set_tooltip_text("username")
+        self.username_entry.set_tooltip_text("server username")
         self.host_entry = Gtk.Entry()
         self.host_entry.set_max_length(128)
         self.host_entry.set_width_chars(24)
         self.host_entry.connect("changed", self.validate)
         self.host_entry.connect("activate", self.connect_clicked)
-        self.host_entry.set_tooltip_text("hostname")
+        self.host_entry.set_tooltip_text("server hostname or IP address")
         self.ssh_port_entry = Gtk.Entry()
         self.ssh_port_entry.set_max_length(5)
         self.ssh_port_entry.set_width_chars(5)
@@ -405,7 +398,7 @@ class ApplicationWindow:
         hbox.set_spacing(20)
         hbox.pack_start(Gtk.Label("Encoding: "))
         self.encoding_combo = Gtk.MenuButton()
-        encodings = ["auto"]+[x for x in PREFERED_ENCODING_ORDER]
+        encodings = ["auto"]+[x for x in PREFERRED_ENCODING_ORDER]
         server_encodings = encodings
         es = make_encodingsmenu(self.get_current_encoding, self.set_new_encoding, encodings, server_encodings)
         self.encoding_combo.set_popup(es)
@@ -436,7 +429,7 @@ class ApplicationWindow:
         # Connect button:
         self.connect_btn = Gtk.Button("Connect")
         self.connect_btn.connect("clicked", self.connect_clicked)
-        connect_icon = self.get_icon("retry.png")
+        connect_icon = get_icon_pixbuf("retry.png")
         if connect_icon:
             self.connect_btn.set_image(scaled_image(connect_icon, 24))
         hbox.pack_start(self.connect_btn)
@@ -451,17 +444,22 @@ class ApplicationWindow:
         theme = Gtk.IconTheme.get_default()
         try:
             pixbuf = theme.load_icon(icon_name, Gtk.IconSize.BUTTON, Gtk.IconLookupFlags.USE_BUILTIN)
-        except:
+        except Exception:
             pixbuf = None
         if not pixbuf:
-            icon_filename = os.path.join(get_icon_dir(), "%s.png" % icon_name)
-            if os.path.exists(icon_filename):
-                pixbuf = GdkPixbuf.Pixbuf.new_from_file(icon_filename)
+            pixbuf = get_icon_pixbuf("%s.png" % icon_name)
             if pixbuf:
                 for size in (16, 32, 48):
                     scaled = pixbuf.scale_simple(size, size, GdkPixbuf.InterpType.BILINEAR)
                     Gtk.IconTheme.add_builtin_icon(icon_name, size, scaled)
-                pixbuf = theme.load_icon(icon_name, Gtk.IconSize.BUTTON, Gtk.IconLookupFlags.USE_BUILTIN)
+                try:
+                    reload = theme.load_icon(icon_name, Gtk.IconSize.BUTTON, Gtk.IconLookupFlags.USE_BUILTIN)
+                except Exception:
+                    log("button: failed to load icon after adding to builtins", exc_info=True)
+                    size = 32
+                    pixbuf = pixbuf.scale_simple(size, size, GdkPixbuf.InterpType.BILINEAR)
+                else:
+                    pixbuf = reload
         if pixbuf:
             image = Gtk.Image.new_from_pixbuf(pixbuf)
             button.add(image)
@@ -473,7 +471,7 @@ class ApplicationWindow:
 
     def accel_close(self, *args):
         log("accel_close%s", args)
-        gtk_main_quit_really()
+        Gtk.main_quit()
 
     def validate(self, *args):
         mode = self.mode_combo.get_active_text().lower()
@@ -553,11 +551,6 @@ class ApplicationWindow:
     def run(self):
         Gtk.main()
 
-    def get_icon(self, icon_name):
-        icon_filename = os.path.join(get_icon_dir(), icon_name)
-        if os.path.exists(icon_filename):
-            return GdkPixbuf.Pixbuf.new_from_file(icon_filename)
-        return None
 
     def mode_changed(self, *_args):
         mode = self.mode_combo.get_active_text().lower()
@@ -784,7 +777,7 @@ class ApplicationWindow:
             self.is_putty = ssh_cmd_0.endswith("plink") or ssh_cmd_0.endswith("plink.exe")
             self.is_paramiko = ssh_cmd_0 =="paramiko"
             full_ssh = ssh_cmd[:]
-            full_ssh += add_ssh_args(username, password, host, self.config.ssh_port, self.is_putty, self.is_paramiko)
+            full_ssh += add_ssh_args(username, password, host, self.config.ssh_port, None, self.is_putty, self.is_paramiko)
             if username:
                 params["username"] = username
             if self.nostrict_host_check.get_active():
@@ -823,9 +816,18 @@ class ApplicationWindow:
         #so update the client now:
         def raise_exception(*args):
             raise Exception(*args)
+        configure_env(self.config.env)
+        configure_logging(self.config, "attach")
+        configure_network(self.config)
         self.client = make_client(raise_exception, self.config)
+        self.client.show_progress(30, "client configuration")
         self.client.init(self.config)
+        self.client.show_progress(40, "loading user interface")
+        self.client.init_ui(self.config)
         self.client.username = username
+        def handshake_complete(*_args):
+            self.client.show_progress(100, "connection established")
+        self.client.after_handshake(handshake_complete)
         self.set_info_text("Connecting...")
         start_thread(self.do_connect_builtin, "connect", daemon=True, args=(params,))
 
@@ -876,7 +878,7 @@ class ApplicationWindow:
             log("do_quit%s", args)
             self.clean_client()
             self.destroy()
-            gtk_main_quit_really()
+            Gtk.main_quit()
 
         def handle_client_quit(exit_launcher=False):
             w = self.window
@@ -919,12 +921,17 @@ class ApplicationWindow:
             self.set_info_text("Running")
             self.window.hide()
         self.client.connect("first-ui-received", first_ui_received)
+        if Gtk.main_level()>0:
+            #no need to start a new main loop:
+            self.client.gtk_main = noop
         try:
             self.client.run()
             log("client.run() returned")
         except Exception as e:
             log.error("client error", exc_info=True)
             self.handle_exception(e)
+        if self.client.gtk_main==noop:
+            return
         log("exit_launcher=%s", self.exit_launcher)
         #if we're using "autoconnect",
         #the main loop was running from here,
@@ -1042,7 +1049,7 @@ class ApplicationWindow:
         self.exit_launcher = True
         self.clean_client()
         self.close_window()
-        gtk_main_quit_really()
+        Gtk.main_quit()
         return False
 
     def update_options_from_URL(self, url):
@@ -1110,7 +1117,7 @@ def exception_dialog(title):
     md.show_all()
     def close_dialog(*_args):
         md.destroy()
-        gtk_main_quit_really()
+        Gtk.main_quit()
     md.connect("response", close_dialog)
     md.connect("close", close_dialog)
     Gtk.main()
@@ -1126,7 +1133,6 @@ def main(argv):
 def do_main(argv):
     from xpra.os_util import SIGNAMES
     from xpra.scripts.main import InitExit, InitInfo
-    from xpra.gtk_common.quit import gtk_main_quit_on_fatal_exceptions_enable
     from xpra.platform.gui import init as gui_init, ready as gui_ready
 
     if POSIX and not OSX:
@@ -1134,7 +1140,6 @@ def do_main(argv):
         init_gdk_display_source()
 
     gui_init()
-    gtk_main_quit_on_fatal_exceptions_enable()
     try:
         from xpra.scripts.parsing import parse_cmdline, fixup_debug_option
         options, args = parse_cmdline(argv)
@@ -1162,6 +1167,8 @@ def do_main(argv):
             client = app.client
             if client:
                 client.cleanup()
+            else:
+                Gtk.main_quit()
             GLib.timeout_add(1000, app.set_info_text, "got signal %s" % SIGNAMES.get(signum, signum))
             GLib.timeout_add(1000, app.set_info_color, True)
         register_os_signals(handle_signal, "Client Launcher")
