@@ -5,7 +5,7 @@
 
 from xpra.util import envbool
 from xpra.net.websockets.common import make_websocket_accept_hash
-from xpra.server.http_handler import HTTPRequestHandler
+from xpra.net.http_handler import HTTPRequestHandler
 from xpra.log import Logger
 
 log = Logger("network", "websocket")
@@ -24,7 +24,7 @@ class WebSocketRequestHandler(HTTPRequestHandler):
 
     def __init__(self, sock, addr, new_websocket_client,
                  web_root="/usr/share/xpra/www/",
-                 http_headers_dir="/usr/share/xpra/http-headers", script_paths=None):
+                 http_headers_dir="/etc/xpra/http-headers", script_paths=None):
         self.new_websocket_client = new_websocket_client
         self.only_upgrade = WEBSOCKET_ONLY_UPGRADE
         super().__init__(sock, addr, web_root, http_headers_dir, script_paths)
@@ -32,6 +32,9 @@ class WebSocketRequestHandler(HTTPRequestHandler):
     def handle_websocket(self):
         log("handle_websocket() calling %s, request=%s (%s)",
             self.new_websocket_client, self.request, type(self.request))
+        log("headers:")
+        for k,v in self.headers.items():
+            log(" %s=%s", k, v)
         ver = self.headers.get('Sec-WebSocket-Version')
         if ver is None:
             raise Exception("Missing Sec-WebSocket-Version header")
