@@ -14,12 +14,19 @@ from xpra.os_util import get_util_logger, osexpand, umask_context
 from xpra.platform.dotxpra_common import PREFIX, LIVE, DEAD, UNKNOWN, INACCESSIBLE
 from xpra.platform import platform_import
 
+DISPLAY_PREFIX = ":"
+
 
 def norm_makepath(dirpath, name):
-    if name[0]==":":
-        name = name[1:]
+    if DISPLAY_PREFIX and name.startswith(DISPLAY_PREFIX):
+        name = name[len(DISPLAY_PREFIX):]
     return os.path.join(dirpath, PREFIX + name)
 
+def strip_display_prefix(s):
+    if s.startswith(DISPLAY_PREFIX):
+        return s[len(DISPLAY_PREFIX):]
+    return s
+
 def debug(msg, *args, **kwargs):
     log = get_util_logger()
     log(msg, *args, **kwargs)
@@ -48,15 +55,29 @@ class DotXpra:
         return "DotXpra(%s, %s - %i:%i - %s)" % (self._sockdir, self._sockdirs, self.uid, self.gid, self.username)
 
     def mksockdir(self, d, mode=0o700, uid=None, gid=None):
-        if d and not os.path.exists(d):
+        if not d:
+            return
+        if not os.path.exists(d):
             if uid is None:
                 uid = self.uid
             if gid is None:
                 gid = self.gid
+            parent = os.path.dirname(d)
+            if parent and parent!="/" and not os.path.exists(parent):
+                self.mksockdir(parent, mode, uid, gid)
             with umask_context(0):
                 os.mkdir(d, mode)
             if uid!=os.getuid() or gid!=os.getgid():
                 os.lchown(d, uid, gid)
+        elif d!="/tmp":
+            try:
+                st_mode = os.stat(d).st_mode
+                if st_mode&0o777!=mode:
+                    log = get_util_logger()
+                    log.warn("Warning: socket directory '%s'", d)
+                    log.warn(" expected permissions %s but found %s", oct(mode), oct(st_mode&0o777))
+            except OSError:
+                get_util_logger().log("mksockdir%s", (d, mode, uid, gid), exc_info=True)
 
     def socket_expand(self, path):
         return osexpand(path, self.username, uid=self.uid, gid=self.gid)
@@ -138,7 +159,7 @@ class DotXpra:
             seen.add(real_dir)
             #ie: "~/.xpra/HOSTNAME-"
             base = os.path.join(d, PREFIX)
-            potential_sockets = glob.glob(base + display.lstrip(":"))
+            potential_sockets = glob.glob(base + strip_display_prefix(display))
             for sockpath in sorted(potential_sockets):
                 try:
                     s = os.stat(sockpath)
@@ -147,7 +168,7 @@ class DotXpra:
                     #socket cannot be accessed
                     continue
                 if stat.S_ISSOCK(s.st_mode):
-                    local_display = ":"+sockpath[len(base):]
+                    local_display = DISPLAY_PREFIX+sockpath[len(base):]
                     if local_display!=display:
                         debug("get_display_state: '%s' display does not match (%s vs %s)",
                               sockpath, local_display, display)
@@ -179,9 +200,10 @@ class DotXpra:
             #ie: "~/.xpra/HOSTNAME-"
             base = os.path.join(d, PREFIX)
             if matching_display:
-                potential_sockets = glob.glob(base + matching_display.lstrip(":"))
+                dstr = strip_display_prefix(matching_display)
             else:
-                potential_sockets = glob.glob(base + "*")
+                dstr = "*"
+            potential_sockets = glob.glob(base + dstr)
             results = []
             for sockpath in sorted(potential_sockets):
                 try:
@@ -200,7 +222,7 @@ class DotXpra:
                     if matching_state and state!=matching_state:
                         debug("socket_details: '%s' state does not match (%s vs %s)", sockpath, state, matching_state)
                         continue
-                    local_display = ":"+sockpath[len(base):]
+                    local_display = DISPLAY_PREFIX+sockpath[len(base):]
                     results.append((state, local_display, sockpath))
             if results:
                 sd[d] = results
@@ -210,4 +232,5 @@ class DotXpra:
 #win32 re-defines DotXpra for namedpipes:
 platform_import(globals(), "dotxpra", False,
                 "DotXpra",
+                "DISPLAY_PREFIX",
                 "norm_makepath")
