@@ -147,7 +147,19 @@ def main(): # pragma: no cover
 
         from gi.repository import Gtk
         menu = Gtk.Menu()
-        item = Gtk.MenuItem(label="Some Menu Item Here")
+        item = Gtk.MenuItem(label="Top Menu Item 1")
+        submenu = Gtk.Menu()
+        item.set_submenu(submenu)
+        sub = Gtk.MenuItem(label="Sub Menu Item 1")
+        subsubmenu = Gtk.Menu()
+        sub.set_submenu(subsubmenu)
+        subsubmenu.append(Gtk.MenuItem(label="Sub Sub Menu Item 1"))
+        subsubmenu.append(Gtk.MenuItem(label="Sub Sub Menu Item 2"))
+        submenu.append(sub)
+        sub = Gtk.MenuItem(label="Sub Menu Item 2")
+        submenu.append(sub)
+        menu.append(item)
+        item = Gtk.MenuItem(label="Top Menu Item 2")
         menu.append(item)
         menu.show_all()
         a = AppindicatorTray(None, None, menu, "test", "xpra.png", None, None, None, Gtk.main_quit)
