@@ -1,30 +1,28 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
 # Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2010-2019 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2020 Antoine Martin <antoine@xpra.org>
 # Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
+#pylint: disable=wrong-import-position
+
 import sys
 import os.path
+import gi
+gi.require_version('Gdk', '3.0')
+gi.require_version('Gtk', '3.0')
+gi.require_version('Pango', '1.0')
+from gi.repository import GLib, Gdk, Gtk
 
 from xpra.util import flatten_dict, envbool
-from xpra.os_util import monotonic_time, register_SIGUSR_signals
-from xpra.gtk_common.gobject_compat import (
-    import_gdk, import_glib, is_gtk3,
-    register_os_signals,
-    )
-from xpra.gtk_common.quit import (
-    gtk_main_quit_really,
-    gtk_main_quit_on_fatal_exceptions_enable,
-    gtk_main_quit_on_fatal_exceptions_disable,
-    )
+from xpra.os_util import monotonic_time
+from xpra.gtk_common.gobject_compat import register_os_signals
 from xpra.server import server_features
 from xpra.server.server_base import ServerBase
 from xpra.gtk_common.gtk_util import (
-    get_gtk_version_info, gtk_main, display_get_default, get_root_size,
-    keymap_get_for_display, icon_theme_get_default, ICON_SIZE_BUTTON,
+    get_gtk_version_info, get_root_size,
     )
 from xpra.log import Logger
 
@@ -35,9 +33,6 @@ screenlog = Logger("server", "screen")
 cursorlog = Logger("server", "cursor")
 notifylog = Logger("notify")
 
-glib = import_glib()
-gdk = import_gdk()
-
 
 class GTKServerBase(ServerBase):
     """
@@ -49,32 +44,32 @@ class GTKServerBase(ServerBase):
 
     def __init__(self):
         log("GTKServerBase.__init__()")
-        self.idle_add = glib.idle_add
-        self.timeout_add = glib.timeout_add
-        self.source_remove = glib.source_remove
+        self.idle_add = GLib.idle_add
+        self.timeout_add = GLib.timeout_add
+        self.source_remove = GLib.source_remove
         self.cursor_suspended = False
         self.ui_watcher = None
         ServerBase.__init__(self)
 
     def watch_keymap_changes(self):
         ### Set up keymap change notification:
-        display = display_get_default()
-        keymap = keymap_get_for_display(display)
+        display = Gdk.Display.get_default()
+        keymap = Gdk.Keymap.get_for_display(display)
         keymap.connect("keys-changed", self._keys_changed)
 
     def install_signal_handlers(self, callback):
-        register_os_signals(callback)
-        register_SIGUSR_signals(glib.idle_add)
-
-    def signal_quit(self, signum, frame=None):
-        gtk_main_quit_on_fatal_exceptions_disable()
-        super(GTKServerBase, self).signal_quit(signum, frame)
+        sstr = "%s server" % self.get_server_mode()
+        register_os_signals(callback, sstr)
+        from xpra.gtk_common.gobject_compat import register_SIGUSR_signals
+        register_SIGUSR_signals(sstr)
 
     def do_quit(self):
-        log("do_quit: calling gtk_main_quit_really")
-        gtk_main_quit_on_fatal_exceptions_disable()
-        gtk_main_quit_really()
-        log("do_quit: gtk_main_quit_really done")
+        log("do_quit: calling Gtk.main_quit")
+        Gtk.main_quit()
+        log("do_quit: Gtk.main_quit done")
+        #from now on, we can't rely on the main loop:
+        from xpra.os_util import register_SIGUSR_signals
+        register_SIGUSR_signals()
 
     def do_cleanup(self):
         ServerBase.do_cleanup(self)
@@ -88,41 +83,40 @@ class GTKServerBase(ServerBase):
         # (if gtk has been loaded)
         gdk_mod = sys.modules.get("gtk.gdk") or sys.modules.get("gi.repository.Gdk")
         #bug 2328: python3 shadow server segfault on Ubuntu 16.04
-        from xpra.os_util import getUbuntuVersion, is_Ubuntu, PYTHON2
-        safe_close = PYTHON2 or self.session_type!="shadow" or not is_Ubuntu() or getUbuntuVersion()>(16, 4)
-        close = envbool("XPRA_CLOSE_GTK_DISPLAY", safe_close)
+        #also crashes on Ubuntu 20.04
+        close = envbool("XPRA_CLOSE_GTK_DISPLAY", False)
+        log("close_gtk_display() close=%s, gdk_mod=%s",
+            close, gdk_mod)
         if close and gdk_mod:
-            if is_gtk3():
-                displays = gdk.DisplayManager.get().list_displays()
-            else:
-                displays = gdk.display_manager_get().list_displays()
+            displays = Gdk.DisplayManager.get().list_displays()
+            log("close_gtk_display() displays=%s", displays)
             for d in displays:
+                log("close_gtk_display() closing %s", d)
                 d.close()
 
 
     def do_run(self):
         if UI_THREAD_WATCHER:
             from xpra.platform.ui_thread_watcher import get_UI_watcher
-            self.ui_watcher = get_UI_watcher(glib.timeout_add, glib.source_remove)
+            self.ui_watcher = get_UI_watcher(GLib.timeout_add, GLib.source_remove)
             self.ui_watcher.start()
         if server_features.windows:
-            display = display_get_default()
+            display = Gdk.Display.get_default()
             i=0
             while i<display.get_n_screens():
                 screen = display.get_screen(i)
                 screen.connect("size-changed", self._screen_size_changed)
                 screen.connect("monitors-changed", self._monitors_changed)
                 i += 1
-        gtk_main_quit_on_fatal_exceptions_enable()
-        log("do_run() calling %s", gtk_main)
-        gtk_main()
+        log("do_run() calling %s", Gtk.main)
+        Gtk.main()
         log("do_run() end of gtk.main()")
 
 
     def make_hello(self, source):
-        capabilities = ServerBase.make_hello(self, source)
+        capabilities = super().make_hello(source)
         if source.wants_display:
-            display = display_get_default()
+            display = Gdk.Display.get_default()
             max_size = tuple(display.get_maximal_cursor_size())
             capabilities.update({
                 "display"               : display.get_name(),
@@ -134,9 +128,9 @@ class GTKServerBase(ServerBase):
         return capabilities
 
     def get_ui_info(self, proto, *args):
-        info = ServerBase.get_ui_info(self, proto, *args)
+        info = super().get_ui_info(proto, *args)
         info.setdefault("server", {}).update({
-                                              "display"             : display_get_default().get_name(),
+                                              "display"             : Gdk.Display.get_default().get_name(),
                                               "root_window_size"    : self.get_root_window_size(),
                                               })
         info.setdefault("cursor", {}).update(self.get_ui_cursor_info())
@@ -168,15 +162,15 @@ class GTKServerBase(ServerBase):
 
     def send_initial_cursors(self, ss, _sharing=False):
         #cursors: get sizes and send:
-        display = display_get_default()
+        display = Gdk.Display.get_default()
         self.cursor_sizes = int(display.get_default_cursor_size()), display.get_maximal_cursor_size()
         cursorlog("send_initial_cursors() cursor_sizes=%s", self.cursor_sizes)
         ss.send_cursor()
 
-    def get_ui_cursor_info(self):
+    def get_ui_cursor_info(self) -> dict:
         #(from UI thread)
         #now cursor size info:
-        display = display_get_default()
+        display = Gdk.Display.get_default()
         pos = display.get_default_screen().get_root_window().get_pointer()[-3:-1]
         cinfo = {"position" : pos}
         for prop, size in {
@@ -190,7 +184,7 @@ class GTKServerBase(ServerBase):
 
     def do_get_info(self, proto, *args):
         start = monotonic_time()
-        info = ServerBase.do_get_info(self, proto, *args)
+        info = super().do_get_info(proto, *args)
         vi = get_gtk_version_info()
         vi["type"] = "Python/gtk"
         info.setdefault("server", {}).update(vi)
@@ -210,7 +204,7 @@ class GTKServerBase(ServerBase):
 
     def calculate_workarea(self, maxw, maxh):
         screenlog("calculate_workarea(%s, %s)", maxw, maxh)
-        workarea = gdk.Rectangle()
+        workarea = Gdk.Rectangle()
         workarea.width = maxw
         workarea.height = maxh
         for ss in self._server_sources.values():
@@ -225,23 +219,22 @@ class GTKServerBase(ServerBase):
                 #display: [':0.0', 2560, 1600, 677, 423, [['DFP2', 0, 0, 2560, 1600, 646, 406]], 0, 0, 2560, 1574]
                 if len(display)>=10:
                     work_x, work_y, work_w, work_h = display[6:10]
-                    display_workarea = gdk.Rectangle()
+                    display_workarea = Gdk.Rectangle()
                     display_workarea.x = work_x
                     display_workarea.y = work_y
                     display_workarea.width = work_w
                     display_workarea.height = work_h
                     screenlog("calculate_workarea() found %s for display %s", display_workarea, display[0])
-                    if is_gtk3():
-                        success, workarea = workarea.intersect(display_workarea)
-                        assert success
-                    else:
-                        workarea = workarea.intersect(display_workarea)
+                    success, workarea = workarea.intersect(display_workarea)
+                    if not success:
+                        log.warn("Warning: failed to calculate workarea")
+                        log.warn(" as intersection of %s and %s", (maxw, maxh), (work_x, work_y, work_w, work_h))
         #sanity checks:
         screenlog("calculate_workarea(%s, %s) workarea=%s", maxw, maxh, workarea)
-        if workarea.width==0 or workarea.height==0:
+        if workarea.width==0 or workarea.height==0 or workarea.width>=32768-8192 or workarea.height>=32768-8192:
             screenlog.warn("Warning: failed to calculate a common workarea")
             screenlog.warn(" using the full display area: %ix%i", maxw, maxh)
-            workarea = gdk.Rectangle()
+            workarea = Gdk.Rectangle()
             workarea.width = maxw
             workarea.height = maxh
         self.set_workarea(workarea)
@@ -258,7 +251,7 @@ class GTKServerBase(ServerBase):
 
     def _move_pointer(self, _wid, pos, *_args):
         x, y = pos
-        display = display_get_default()
+        display = Gdk.Display.get_default()
         display.warp_pointer(display.get_default_screen(), x, y)
 
     def do_process_button_action(self, *args):
@@ -288,14 +281,22 @@ class GTKServerBase(ServerBase):
         from PIL import Image
         if os.path.isabs(icon_string):
             if os.path.exists(icon_string) and os.path.isfile(icon_string):
-                img = Image.open(icon_string)
-                w, h = img.size
+                #should we be using pillow.decoder.open_only here? meh
+                try:
+                    img = Image.open(icon_string)
+                except Exception as e:
+                    log("%s(%s)", Image.open, icon_string)
+                    log.warn("Warning: unable to load notification icon file")
+                    log.warn(" '%s'", icon_string)
+                    log.warn(" %s", e)
+                else:
+                    w, h = img.size
         else:
             #try to find it in the theme:
-            theme = icon_theme_get_default()
+            theme = Gtk.IconTheme.get_default()
             if theme:
                 try:
-                    icon = theme.load_icon(icon_string, ICON_SIZE_BUTTON, 0)
+                    icon = theme.load_icon(icon_string, Gtk.IconSize.BUTTON, 0)
                 except Exception as e:
                     notifylog("failed to load icon '%s' from default theme: %s", icon_string, e)
                 else:
