@@ -1,23 +1,26 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2010-2018 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2010-2020 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
+#pylint: disable-msg=E1101
 
 import os.path
 
+from xpra.util import parse_scaling_value, csv, from0to100
+from xpra.os_util import load_binary_file, bytestostr, strtobytes
+from xpra.simple_stats import std_unit
+from xpra.scripts.config import parse_bool, FALSE_OPTIONS, TRUE_OPTIONS
+from xpra.server.control_command import ArgsControlCommand, ControlError
+from xpra.server.mixins.stub_server_mixin import StubServerMixin
 from xpra.log import Logger
+
 log = Logger("command")
 
-from xpra.util import parse_scaling_value, from0to100
-from xpra.server.control_command import ArgsControlCommand, ControlError
-from xpra.os_util import load_binary_file
-from xpra.util import csv
-from xpra.scripts.config import parse_bool, FALSE_OPTIONS, TRUE_OPTIONS
-from xpra.server.mixins.stub_server_mixin import StubServerMixin
-
-TOGGLE_FEATURES = ("bell", "randr", "cursors", "notifications", "dbus-proxy", "clipboard",
-                   "start-new-commands", "client-shutdown", "webcam", )
+TOGGLE_FEATURES = (
+    "bell", "randr", "cursors", "notifications", "dbus-proxy", "clipboard",
+    "start-new-commands", "client-shutdown", "webcam",
+    )
 
 
 """
@@ -33,10 +36,9 @@ class ServerBaseControlCommands(StubServerMixin):
         def parse_boolean_value(v):
             if str(v).lower() in TRUE_OPTIONS:
                 return True
-            elif str(v).lower() in FALSE_OPTIONS:
+            if str(v).lower() in FALSE_OPTIONS:
                 return False
-            else:
-                raise ControlError("a boolean is required, not %s" % v)
+            raise ControlError("a boolean is required, not %s" % v)
         def parse_4intlist(v):
             if not v:
                 return []
@@ -56,6 +58,8 @@ class ServerBaseControlCommands(StubServerMixin):
 
         for cmd in (
             ArgsControlCommand("focus",                 "give focus to the window id",      validation=[int]),
+            ArgsControlCommand("map",                   "maps the window id",               validation=[int]),
+            ArgsControlCommand("unmap",                 "unmaps the window id",             validation=[int]),
             #window source:
             ArgsControlCommand("suspend",               "suspend screen updates",           max_args=0),
             ArgsControlCommand("resume",                "resume screen updates",            max_args=0),
@@ -75,8 +79,10 @@ class ServerBaseControlCommands(StubServerMixin):
             ArgsControlCommand("compression",           "sets the packet compressor",       min_args=1, max_args=1),
             ArgsControlCommand("encoder",               "sets the packet encoder",          min_args=1, max_args=1),
             ArgsControlCommand("clipboard-direction",   "restrict clipboard transfers",     min_args=1, max_args=1),
+            ArgsControlCommand("clipboard-limits",      "restrict clipboard transfers size", min_args=2, max_args=2, validation=[int, int]),
             ArgsControlCommand("set-lock",              "modify the lock attribute",        min_args=1, max_args=1),
             ArgsControlCommand("set-sharing",           "modify the sharing attribute",     min_args=1, max_args=1),
+            ArgsControlCommand("set-ui-driver",         "set the client connection driving the session", min_args=1, max_args=1),
             #session and clients:
             ArgsControlCommand("client",                "forwards a control command to the client(s)", min_args=1),
             ArgsControlCommand("client-property",       "set a client property",            min_args=4, max_args=5, validation=[int]),
@@ -85,11 +91,14 @@ class ServerBaseControlCommands(StubServerMixin):
             ArgsControlCommand("sound-output",          "control sound forwarding",         min_args=1, max_args=2),
             #windows:
             ArgsControlCommand("workspace",             "move a window to a different workspace", min_args=2, max_args=2, validation=[int, int]),
+            ArgsControlCommand("move",                  "move a window",                    min_args=3, max_args=3, validation=[int, int, int]),
+            ArgsControlCommand("resize",                "resize a window",                  min_args=3, max_args=3, validation=[int, int, int]),
+            ArgsControlCommand("moveresize",            "move and resize a window",         min_args=5, max_args=5, validation=[int, int, int, int, int]),
             ArgsControlCommand("scaling-control",       "set the scaling-control aggressiveness (from 0 to 100)", min_args=1, validation=[from0to100]),
             ArgsControlCommand("scaling",               "set a specific scaling value",     min_args=1, validation=[parse_scaling_value]),
             ArgsControlCommand("auto-refresh",          "set a specific auto-refresh value", min_args=1, validation=[float]),
             ArgsControlCommand("refresh",               "refresh some or all windows",      min_args=0),
-            ArgsControlCommand("encoding",              "picture encoding",                 min_args=1, max_args=1),
+            ArgsControlCommand("encoding",              "picture encoding",                 min_args=2),
             ArgsControlCommand("video-region-enabled",  "enable video region",              min_args=2, max_args=2, validation=[int, parse_boolean_value]),
             ArgsControlCommand("video-region-detection","enable video detection",           min_args=2, max_args=2, validation=[int, parse_boolean_value]),
             ArgsControlCommand("video-region-exclusion-zones","set window regions to exclude from video regions: 'WID,(x,y,w,h),(x,y,w,h),..', ie: '1 (0,10,100,20),(200,300,20,20)'",  min_args=2, max_args=2, validation=[int, parse_4intlist]),
@@ -97,13 +106,15 @@ class ServerBaseControlCommands(StubServerMixin):
             ArgsControlCommand("reset-video-region",    "reset video region heuristics",    min_args=1, max_args=1, validation=[int]),
             ArgsControlCommand("lock-batch-delay",      "set a specific batch delay for a window",       min_args=2, max_args=2, validation=[int, int]),
             ArgsControlCommand("unlock-batch-delay",    "let the heuristics calculate the batch delay again for a window (following a 'lock-batch-delay')",  min_args=1, max_args=1, validation=[int]),
+            ArgsControlCommand("remove-window-filters", "remove all window filters",        min_args=0, max_args=0),
+            ArgsControlCommand("add-window-filter",     "add a window filter",              min_args=4, max_args=5),
             ):
             cmd.do_run = getattr(self, "control_command_%s" % cmd.name.replace("-", "_"))
             self.control_commands[cmd.name] = cmd
         #encoding bits:
         for name in ("quality", "min-quality", "speed", "min-speed"):
             fn = getattr(self, "control_command_%s" % name.replace("-", "_"))
-            self.control_commands[name] = ArgsControlCommand(name, "set encoding %s (from 0 to 100)" % name, run=fn, min_args=1, max_args=1, validation=[from0to100])
+            self.control_commands[name] = ArgsControlCommand(name, "set encoding %s (from 0 to 100)" % name, run=fn, min_args=1, validation=[from0to100])
 
 
     #########################################
@@ -116,6 +127,44 @@ class ServerBaseControlCommands(StubServerMixin):
         self._focus(None, wid, None)
         return "gave focus to window %s" % wid
 
+    def control_command_map(self, wid):
+        if self.readonly:
+            return
+        assert type(wid)==int, "argument should have been an int, but found %s" % type(wid)
+        window = self._id_to_window.get(wid)
+        assert window, "window %i not found" % wid
+        if window.is_tray():
+            return "cannot map tray window %s" % wid
+        if window.is_OR():
+            return "cannot map override redirect window %s" % wid
+        dm = getattr(self, "_desktop_manager", None)
+        assert dm, "%r does not have a desktop manager" % self
+        dm.show_window(window)
+        #window.set_owner(dm)
+        #iconic = window.get_property("iconic")
+        #if iconic:
+        #    window.set_property("iconic", False)
+        #w, h = window.get_geometry()[2:4]
+        #self.refresh_window_area(window, 0, 0, w, h)
+        self.repaint_root_overlay()
+        return "mapped window %s" % wid
+
+    def control_command_unmap(self, wid):
+        if self.readonly:
+            return
+        assert type(wid)==int, "argument should have been an int, but found %s" % type(wid)
+        window = self._id_to_window.get(wid)
+        assert window, "window %i not found" % wid
+        if window.is_tray():
+            return "cannot map tray window %s" % wid
+        if window.is_OR():
+            return "cannot map override redirect window %s" % wid
+        dm = getattr(self, "_desktop_manager", None)
+        assert dm, "%r does not have a desktop manager" % self
+        dm.hide_window(window)
+        self.repaint_root_overlay()
+        return "unmapped window %s" % wid
+
     def control_command_suspend(self):
         for csource in tuple(self._server_sources.values()):
             csource.suspend(True, self._id_to_window)
@@ -180,14 +229,15 @@ class ServerBaseControlCommands(StubServerMixin):
         if client_uuids_str=="UI":
             sources = [ss for ss in self._server_sources.values() if ss.ui_client]
             client_uuids = [ss.uuid for ss in sources]
-            notfound = []
+            notfound = ()
         elif client_uuids_str=="*":
             sources = self._server_sources.values()
             client_uuids = [ss.uuid for ss in sources]
         else:
             client_uuids = client_uuids_str.split(",")
             sources = [ss for ss in self._server_sources.values() if ss.uuid in client_uuids]
-            notfound = [x for x in client_uuids if x not in [ss.uuid for ss in sources]]
+            uuids = tuple(ss.uuid for ss in sources)
+            notfound = any(x for x in client_uuids if x not in uuids)
             if notfound:
                 log.warn("client connection not found for uuid(s): %s", notfound)
         return sources
@@ -198,7 +248,8 @@ class ServerBaseControlCommands(StubServerMixin):
             log(msg)
             return msg
         sources = self._control_get_sources(client_uuids)
-        log("control_command_send_notification(%i, %s, %s, %s) will send to sources %s (matching %s)", nid, title, message, client_uuids, sources, client_uuids)
+        log("control_command_send_notification(%i, %s, %s, %s) will send to sources %s (matching %s)",
+            nid, title, message, client_uuids, sources, client_uuids)
         count = 0
         for source in sources:
             if source.notify(0, nid, "control channel", 0, "", title, message, [], {}, 10, ""):
@@ -228,11 +279,19 @@ class ServerBaseControlCommands(StubServerMixin):
             raise ControlError("no clients found matching: %s" % client_uuids)
         clients = 0
         for ss in sources:
-            if ss.send_open_url(url):
-                clients += 1
+            if hasattr(ss, "send_open_url"):
+                if ss.send_open_url(url):
+                    clients += 1
         return "url sent to %i clients" % clients
 
     def control_command_send_file(self, filename, openit="open", client_uuids="*", maxbitrate=0):
+        #we always get the values as strings from the command interface,
+        #but those may actually be utf8 encoded binary strings,
+        #so we may have to do an ugly roundtrip:
+        try:
+            filename = strtobytes(filename).decode("utf8")
+        except Exception:
+            filename = bytestostr(filename)
         openit = str(openit).lower() in ("open", "true", "1")
         return self.do_control_file_command("send file", client_uuids, filename, "file_transfer", (False, openit))
 
@@ -251,6 +310,11 @@ class ServerBaseControlCommands(StubServerMixin):
         sources = self._control_get_sources(client_uuids)
         if not sources:
             raise ControlError("no clients found matching: %s" % client_uuids)
+        def checksize(file_size):
+            if file_size>self.file_transfer.file_size_limit:
+                raise ControlError("file '%s' is too large: %sB (limit is %sB)" % (
+                    filename, std_unit(file_size), std_unit(self.file_transfer.file_size_limit)))
+
         #find the file and load it:
         actual_filename = os.path.abspath(os.path.expanduser(filename))
         try:
@@ -258,28 +322,51 @@ class ServerBaseControlCommands(StubServerMixin):
             log("os.stat(%s)=%s", actual_filename, stat)
         except os.error:
             log("os.stat(%s)", actual_filename, exc_info=True)
+        else:
+            checksize(stat.st_size)
         if not os.path.exists(actual_filename):
             raise ControlError("file '%s' does not exist" % filename)
         data = load_binary_file(actual_filename)
+        if data is None:
+            raise ControlError("failed to load '%s'" % actual_filename)
         #verify size:
         file_size = len(data)
-        file_size_MB = file_size//1024//1024
-        if file_size_MB>self.file_transfer.file_size_limit:
-            raise ControlError("file '%s' is too large: %iMB (limit is %iMB)" % (filename, file_size_MB, self.file_transfer.file_size_limit))
+        checksize(file_size)
         #send it to each client:
         for ss in sources:
             #ie: ServerSource.file_transfer (found in FileTransferAttributes)
-            if not getattr(ss, source_flag_name):
+            if not getattr(ss, source_flag_name, False):
                 log.warn("Warning: cannot %s '%s'", command_type, filename)
                 log.warn(" client %s does not support this feature", ss)
-            elif file_size_MB>ss.file_size_limit:
+            elif file_size>ss.file_size_limit:
                 log.warn("Warning: cannot %s '%s'", command_type, filename)
-                log.warn(" client %s file size limit is %iMB (file is %iMB)", ss, ss.file_size_limit, file_size_MB)
+                log.warn(" client %s file size limit is %sB (file is %sB)",
+                         ss, std_unit(ss.file_size_limit), std_unit(file_size))
             else:
                 ss.send_file(filename, "", data, file_size, *send_file_args)
         return "%s of '%s' to %s initiated" % (command_type, filename, client_uuids)
 
 
+    def control_command_remove_window_filters(self):
+        #modify the existing list object,
+        #which is referenced by all the sources
+        l = len(self.window_filters)
+        self.window_filters[:] = []
+        return "removed %i window-filters" % l
+
+    def control_command_add_window_filter(self, object_name, property_name, operator, value, client_uuids=""):
+        from xpra.server.window import filters
+        window_filter = filters.get_window_filter(object_name, property_name, operator, value)
+        #log("%s%s=%s", filters.get_window_filter, (object_name, property_name, operator, value), window_filter)
+        if client_uuids=="*":
+            #applies to all sources:
+            self.window_filters.append(("*", window_filter))
+        else:
+            for client_uuid in client_uuids.split(","):
+                self.window_filters.append((client_uuid, window_filter))
+        return "added window-filter: %s for client uuids=%s" % (window_filter, client_uuids)
+
+
     def control_command_compression(self, compression):
         c = compression.lower()
         from xpra.net import compression
@@ -306,7 +393,7 @@ class ServerBaseControlCommands(StubServerMixin):
     def all_send_client_command(self, *client_command):
         """ forwards the command to all clients """
         for source in tuple(self._server_sources.values()):
-            """ forwards to *the* client, if there is *one* """
+            # forwards to *the* client, if there is *one*
             if client_command[0] not in source.control_commands:
                 log.info("client command '%s' not forwarded to client %s (not supported)", client_command, source)
             else:
@@ -314,7 +401,7 @@ class ServerBaseControlCommands(StubServerMixin):
 
     def control_command_client(self, *args):
         self.all_send_client_command(*args)
-        return "client control command '%s' forwarded to clients" % str(args)
+        return "client control command %s forwarded to clients" % str(args)
 
     def control_command_client_property(self, wid, uuid, prop, value, conv=None):
         wid = int(wid)
@@ -334,6 +421,7 @@ class ServerBaseControlCommands(StubServerMixin):
         log.info("changed session name: %s", self.session_name)
         #self.all_send_client_command("name", name)    not supported by any clients, don't bother!
         self.setting_changed("session_name", name)
+        self.mdns_update()
         return "session name set to %s" % name
 
     def _control_windowsources_from_args(self, *args):
@@ -348,7 +436,7 @@ class ServerBaseControlCommands(StubServerMixin):
                 try:
                     wid = int(x)
                 except:
-                    raise ControlError("invalid window id: %s" % x)
+                    raise ControlError("invalid window id: %s" % x) from None
                 if wid in self._id_to_window:
                     wids.append(wid)
                 else:
@@ -392,15 +480,15 @@ class ServerBaseControlCommands(StubServerMixin):
         return "refreshed windows %s" % str(wids)
 
     def control_command_scaling_control(self, scaling_control, *wids):
-        for ws, window in self._control_windowsources_from_args(*wids).items():
+        for ws in tuple(self._control_windowsources_from_args(*wids).keys()):
             ws.set_scaling_control(scaling_control)
-            ws.refresh(window)
+            ws.refresh()
         return "scaling-control set to %s on windows %s" % (scaling_control, wids)
 
     def control_command_scaling(self, scaling, *wids):
-        for ws, window in self._control_windowsources_from_args(*wids).items():
+        for ws in tuple(self._control_windowsources_from_args(*wids).keys()):
             ws.set_scaling(scaling)
-            ws.refresh(window)
+            ws.refresh()
         return "scaling set to %s on windows %s" % (str(scaling), wids)
 
     def control_command_encoding(self, encoding, *args):
@@ -410,9 +498,9 @@ class ServerBaseControlCommands(StubServerMixin):
             strict = args[0]=="strict"
             args = args[1:]
         wids = args
-        for ws, window in self._control_windowsources_from_args(*wids).items():
+        for ws in tuple(self._control_windowsources_from_args(*wids).keys()):
             ws.set_new_encoding(encoding, strict)
-            ws.refresh(window, {})
+            ws.refresh()
         return "set encoding to %s%s for windows %s" % (encoding, ["", " (strict)"][int(strict or 0)], wids)
 
     def control_command_clipboard_direction(self, direction, *_args):
@@ -424,10 +512,19 @@ class ServerBaseControlCommands(StubServerMixin):
         self.clipboard_direction = direction
         can_send = direction in ("to-server", "both")
         can_receive = direction in ("to-client", "both")
-        self._clipboard_helper.set_direction(can_send, can_receive)
+        ch.set_direction(can_send, can_receive)
         msg = "clipboard direction set to '%s'" % direction
         log(msg)
-        self.setting_changed("clipboard_direction", direction)
+        self.setting_changed("clipboard-direction", direction)
+        return msg
+
+    def control_command_clipboard_limits(self, max_send, max_recv, *_args):
+        ch = self._clipboard_helper
+        assert self.clipboard and ch
+        ch.set_limits(max_send, max_recv)
+        msg = "clipboard send limit set to %d, recv limit set to %d (single copy/paste)" % (max_send, max_recv)
+        log(msg)
+        self.setting_changed("clipboard-limits", {'send': max_send, 'recv': max_recv})
         return msg
 
     def _control_video_subregions_from_wid(self, wid):
@@ -490,6 +587,15 @@ class ServerBaseControlCommands(StubServerMixin):
         self.setting_changed("sharing-toggle", sharing is None)
         return "sharing set to %s" % self.sharing
 
+    def control_command_set_ui_driver(self, uuid):
+        ss = [s for s in self._server_sources.values() if s.uuid==uuid]
+        if not ss:
+            return "source not found for uuid '%s'" % uuid
+        elif len(ss)>1:
+            return "more than one source found for uuid '%s'" % uuid
+        else:
+            self.set_ui_driver(ss)
+            return "ui-driver set to %s" % ss
 
     def control_command_key(self, keycode_str, press = True):
         if self.readonly:
@@ -501,14 +607,15 @@ class ServerBaseControlCommands(StubServerMixin):
                 keycode = int(keycode_str)
             assert keycode>0 and keycode<=255
         except:
-            raise ControlError("invalid keycode specified: '%s' (must be a number between 1 and 255)" % keycode_str)
+            raise ControlError("invalid keycode specified: '%s' (must be a number between 1 and 255)" % keycode_str) from None
         if press is not True:
             if press in ("1", "press"):
                 press = True
             elif press in ("0", "unpress"):
                 press = False
             else:
-                raise ControlError("if present, the press argument must be one of: %s", ("1", "press", "0", "unpress"))
+                raise ControlError("if present, the press argument must be one of: %s" %
+                                   csv(("1", "press", "0", "unpress")))
         self.fake_key(keycode, press)
 
     def control_command_sound_output(self, *args):
@@ -520,15 +627,53 @@ class ServerBaseControlCommands(StubServerMixin):
     def control_command_workspace(self, wid, workspace):
         window = self._id_to_window.get(wid)
         if not window:
-            raise ControlError("window %s does not exist", wid)
+            raise ControlError("window %s does not exist" % wid)
         if "workspace" not in window.get_property_names():
-            raise ControlError("cannot set workspace on window %s", window)
+            raise ControlError("cannot set workspace on window %s" % window)
         if workspace<0:
-            raise ControlError("invalid workspace value: %s", workspace)
+            raise ControlError("invalid workspace value: %s" % workspace)
         window.set_property("workspace", workspace)
         return "window %s moved to workspace %s" % (wid, workspace)
 
 
+    def control_command_move(self, wid, x, y):
+        window = self._id_to_window.get(wid)
+        if not window:
+            raise ControlError("window %s does not exist" % wid)
+        ww, wh = window.get_dimensions()
+        count = 0
+        for source in tuple(self._server_sources.values()):
+            move_resize_window = getattr(source, "move_resize_window", None)
+            if move_resize_window:
+                move_resize_window(wid, window, x, y, ww, wh)
+                count += 1
+        return "window %s moved to %i,%i for %i clients" % (wid, x, y, count)
+
+    def control_command_resize(self, wid, w, h):
+        window = self._id_to_window.get(wid)
+        if not window:
+            raise ControlError("window %s does not exist" % wid)
+        count = 0
+        for source in tuple(self._server_sources.values()):
+            resize_window = getattr(source, "resize_window", None)
+            if resize_window:
+                resize_window(wid, window, w, h)
+                count += 1
+        return "window %s resized to %ix%i for %i clients" % (wid, w, h, count)
+
+    def control_command_moveresize(self, wid, x, y, w, h):
+        window = self._id_to_window.get(wid)
+        if not window:
+            raise ControlError("window %s does not exist" % wid)
+        count = 0
+        for source in tuple(self._server_sources.values()):
+            move_resize_window = getattr(source, "move_resize_window", None)
+            if move_resize_window:
+                move_resize_window(wid, window, x, y, w, h)
+                count += 1
+        return "window %s moved to %i,%i and resized to %ix%i for %i clients" % (wid, x, y, w, h, count)
+
+
     def _process_command_request(self, _proto, packet):
         """ client sent a command request through its normal channel """
         assert len(packet)>=2, "invalid command request packet (too small!)"
