@@ -1,21 +1,28 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2010-2018 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2010-2020 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
-import os
+from xpra.util import typedict
+from xpra.server.source.stub_source_mixin import StubSourceMixin
 
 from xpra.log import Logger
 log = Logger("mmap")
 
-from xpra.os_util import WIN32
-from xpra.simple_stats import std_unit
-from xpra.server.source.stub_source_mixin import StubSourceMixin
-
 
 class MMAP_Connection(StubSourceMixin):
 
+    @classmethod
+    def is_needed(cls, caps : typedict) -> bool:
+        #pre 2.3 clients;
+        if caps.strget("mmap_file"):
+            return True
+        v = caps.rawget("mmap")
+        #we should be receiving a dict with mmap attributes
+        #(but pre v4 clients also send a boolean telling us if mmap is supported by the platform..)
+        return isinstance(v, dict)
+
     def __init__(self):
         self.supports_mmap = False
         self.mmap_filename = None
@@ -42,9 +49,11 @@ class MMAP_Connection(StubSourceMixin):
             mmap.close()
 
 
-    def parse_client_caps(self, c):
+    def parse_client_caps(self, c : typedict):
+        import os
+        from xpra.os_util import WIN32
         self.mmap_client_namespace = c.boolget("mmap.namespace", False)
-        sep = ["_", "."][self.mmap_client_namespace] 
+        sep = "." if self.mmap_client_namespace else "_"
         def mmapattr(k):
             return "mmap%s%s" % (sep, k)
         mmap_filename = c.strget(mmapattr("file"))
@@ -55,8 +64,13 @@ class MMAP_Connection(StubSourceMixin):
         mmap_token = c.intget(mmapattr("token"))
         log("mmap supported=%s, token=%s", self.supports_mmap, mmap_token)
         if self.mmap_filename:
-            log("using global server specified mmap file path: '%s'", self.mmap_filename)
-            mmap_filename = self.mmap_filename
+            if os.path.isdir(self.mmap_filename):
+                #use the client's filename, but at the server path:
+                mmap_filename = os.path.join(self.mmap_filename, os.path.basename(mmap_filename))
+                log("using global server specified mmap directory: '%s'", self.mmap_filename)
+            else:
+                log("using global server specified mmap file path: '%s'", self.mmap_filename)
+                mmap_filename = self.mmap_filename
         if not self.supports_mmap:
             log("client enabled mmap but mmap mode is not supported", mmap_filename)
         elif WIN32 and mmap_filename.startswith("/"):
@@ -64,9 +78,15 @@ class MMAP_Connection(StubSourceMixin):
         elif not os.path.exists(mmap_filename):
             log("mmap_file '%s' cannot be found!", mmap_filename)
         else:
-            from xpra.net.mmap_pipe import init_server_mmap, read_mmap_token, write_mmap_token, DEFAULT_TOKEN_INDEX, DEFAULT_TOKEN_BYTES
+            from xpra.net.mmap_pipe import (
+                init_server_mmap,
+                read_mmap_token,
+                write_mmap_token,
+                DEFAULT_TOKEN_INDEX, DEFAULT_TOKEN_BYTES,
+                )
             self.mmap, self.mmap_size = init_server_mmap(mmap_filename, mmap_size)
-            log("found client mmap area: %s, %i bytes - min mmap size=%i", self.mmap, self.mmap_size, self.min_mmap_size)
+            log("found client mmap area: %s, %i bytes - min mmap size=%i in '%s'",
+                self.mmap, self.mmap_size, self.min_mmap_size, mmap_filename)
             if self.mmap_size>0:
                 index = c.intget(mmapattr("token_index"), DEFAULT_TOKEN_INDEX)
                 count = c.intget(mmapattr("token_bytes"), DEFAULT_TOKEN_BYTES)
@@ -80,7 +100,8 @@ class MMAP_Connection(StubSourceMixin):
                     self.mmap_size = 0
                 elif self.mmap_size<self.min_mmap_size:
                     log.warn("Warning: client supplied mmap area is too small, discarding it")
-                    log.warn(" we need at least %iMB and this area is %iMB", self.min_mmap_size//1024//1024, self.mmap_size//1024//1024)
+                    log.warn(" we need at least %iMB and this area is %iMB",
+                             self.min_mmap_size//1024//1024, self.mmap_size//1024//1024)
                     self.mmap.close()
                     self.mmap = None
                     self.mmap_size = 0
@@ -95,14 +116,18 @@ class MMAP_Connection(StubSourceMixin):
                     else:
                         #use the expected default for older versions:
                         self.mmap_client_token_index = DEFAULT_TOKEN_INDEX
-                    write_mmap_token(self.mmap, self.mmap_client_token, self.mmap_client_token_index, self.mmap_client_token_bytes)
+                    write_mmap_token(self.mmap,
+                                     self.mmap_client_token,
+                                     self.mmap_client_token_index,
+                                     self.mmap_client_token_bytes)
         if self.mmap_size>0:
+            from xpra.simple_stats import std_unit
             log.info(" mmap is enabled using %sB area in %s", std_unit(self.mmap_size, unit=1024), mmap_filename)
 
-    def get_caps(self):
+    def get_caps(self) -> dict:
         caps = {"mmap_enabled" : self.mmap_size>0}
         if self.mmap_client_token:
-            sep = ["_", "."][self.mmap_client_namespace]
+            sep = "." if self.mmap_client_namespace else "_"
             def mmapattr(name, value):
                 caps["mmap%s%s" % (sep, name)] = value
             mmapattr("token",       self.mmap_client_token)
@@ -110,7 +135,7 @@ class MMAP_Connection(StubSourceMixin):
             mmapattr("token_bytes", self.mmap_client_token_bytes)
         return caps
 
-    def get_info(self):
+    def get_info(self) -> dict:
         return {
             "mmap" : {
                 "supported"     : self.supports_mmap,
